// file      : xsd/cxx/elements.hxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2006 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#ifndef CXX_ELEMENTS_HXX
#define CXX_ELEMENTS_HXX

#include <cult/types.hxx>
#include <cult/containers/deque.hxx>
#include <cult/containers/vector.hxx>

#include <xsd-frontend/semantic-graph.hxx>
#include <xsd-frontend/traversal.hxx>

#include <elements.hxx>

#include <locale>
#include <ostream>

namespace CXX
{
  using std::endl;
  typedef WideString String;


  // On some platforms std::toupper can be something other than a
  // function with C++ linkage.
  //
  wchar_t
  upcase (wchar_t c);


  // Exceptions.
  //

  struct NoNamespaceMapping
  {
    NoNamespaceMapping (SemanticGraph::Path const& file,
                        UnsignedLong line,
                        UnsignedLong column,
                        String const& ns)
        : file_ (file),
          line_ (line),
          column_ (column),
          ns_ (ns)
    {
    }


    SemanticGraph::Path const&
    file () const
    {
      return file_;
    }

    UnsignedLong
    line () const
    {
      return line_;
    }

    UnsignedLong
    column () const
    {
      return column_;
    }

    String const&
    ns () const
    {
      return ns_;
    }

  private:
    SemanticGraph::Path file_;
    UnsignedLong line_;
    UnsignedLong column_;
    String ns_;
  };


  //
  //
  class Context
  {
  public:
    typedef Cult::Containers::Vector<String> NamespaceMapping;

  public:
    Context (std::wostream& o,
             NarrowString const& char_type__,
             Boolean include_with_brackets__,
             NarrowString const& include_prefix__,
             NarrowString const& esymbol__,
             Containers::Vector<NarrowString> const& nsm,
             Boolean trace_namespace_regex_,
             Boolean inline_)
        : os (o),
          scope (scope_),
          inner_name (inner_name_),
          char_type (char_type_),
          L (L_),
          string_type (string_type_),
          include_with_brackets (include_with_brackets_),
          include_prefix (include_prefix_),
          esymbol (esymbol_),
          inl (inl_),
          char_type_ (char_type__),
          L_ (char_type == L"wchar_t" ? L"L" : L""),
          string_type_ (L"::std::basic_string< " + char_type + L" >"),
          include_with_brackets_ (include_with_brackets__),
          include_prefix_ (include_prefix__),
          esymbol_ (esymbol__),
          inl_ (inline_ ? L"inline\n" : L""),
          scope_stack (scope_stack_),
          trace_namespace_regex (trace_namespace_regex_),
          ns_mapping (ns_mapping_)
    {
      // Default mapping.
      //
      ns_mapping_.push_back (
        L"#^.* (.*?/)??"L"(([a-zA-Z_]\\w*)(/[a-zA-Z_]\\w*)*)$#$2#");
      ns_mapping_.push_back (
        L"#^XMLSchema.xsd http://www.w3.org/2001/XMLSchema$#xml_schema#");

      // Custom mappings.
      //
      for (Containers::Vector<NarrowString>::ConstIterator
             i (nsm.begin ()), e (nsm.end ()); i != e; ++i)
      {
        ns_mapping_.push_back (*i);
      }
    }

  protected:
    Context (Context& c)
        : os (c.os),
          scope (c.scope),
          inner_name (c.inner_name),
          char_type (c.char_type),
          L (c.L),
          string_type (c.string_type),
          include_with_brackets (c.include_with_brackets),
          include_prefix (c.include_prefix),
          esymbol (c.esymbol),
          inl (c.inl),
          scope_stack (c.scope_stack),
          trace_namespace_regex (c.trace_namespace_regex),
          ns_mapping (c.ns_mapping)
    {
    }

    Context (Context& c, std::wostream& o)
        : os (o),
          scope (c.scope),
          inner_name (c.inner_name),
          char_type (c.char_type),
          L (c.L),
          string_type (c.string_type),
          include_with_brackets (c.include_with_brackets),
          include_prefix (c.include_prefix),
          esymbol (c.esymbol),
          inl (c.inl),
          scope_stack (c.scope_stack),
          trace_namespace_regex (c.trace_namespace_regex),
          ns_mapping (c.ns_mapping)
    {
    }


  public:
    static String
    unclash (String const& name, String const& new_name)
    {
      return name == new_name ? (new_name + L'_') : new_name;
    }

  public:
    Void
    enter_scope (String const& name)
    {
      scope_stack.push_back (name);
      update_scope ();
    }

    Void
    leave_scope ()
    {
      scope_stack.pop_back ();
      update_scope ();
    }

  private:
    Void
    update_scope ()
    {
      scope.clear ();

      Boolean first (true);

      for (ScopeStack::Iterator i (scope_stack.begin ());
           i != scope_stack.end ();
           ++i)
      {
        // We only qualify names until the namespace level.
        //
        if (first)
          first = false;
        else
          scope += L"::";

        scope += *i;
      }

      inner_name.clear ();

      if (!scope_stack.empty ())
        inner_name = scope_stack.back ();
    }

  public:

    // Escape C++ keywords and illegal characters.
    //
    static String
    escape (String const& name);

    // Translate XML namespace name to a C++ identifier.
    //
    String
    ns_name (SemanticGraph::Namespace& ns);

    //
    //
    SemanticGraph::Namespace&
    namespace_ (SemanticGraph::Nameable& n);

    // Original schema namespace name.
    //
    String
    xs_ns_name (SemanticGraph::Nameable& ns);


    // Fully-qualified C++ name.
    //
    String
    fq_name (SemanticGraph::Nameable& n);

  public:
    static Boolean
    skip (SemanticGraph::Member& m)
    {
      return !m.scope ().is_a<SemanticGraph::Namespace> () &&
        m.context ().count ("min") == 0;
    }

    static UnsignedLong
    min (SemanticGraph::Member const& m)
    {
      return m.context ().get<UnsignedLong> ("min");
    }

    static UnsignedLong
    max (SemanticGraph::Member const& m)
    {
      return m.context ().get<UnsignedLong> ("max");
    }

  public:
    // Get escaped name.
    //
    static String const&
    ename (SemanticGraph::Nameable const& n)
    {
      return n.context ().get<String> ("name");
    }

  public:
    std::wostream& os;
    String& scope;
    String& inner_name;

    String& char_type;
    String& L;                  // string literal prefix
    String& string_type;

    Boolean& include_with_brackets;
    String& include_prefix;

    String& esymbol;
    String& inl;

    static std::locale locale;

  private:
    String scope_;
    String inner_name_;

    String char_type_;
    String L_;
    String string_type_;

    Boolean include_with_brackets_;
    String include_prefix_;

    String esymbol_;
    String inl_;

    typedef
    Cult::Containers::Deque<String>
    ScopeStack;

    ScopeStack& scope_stack;
    ScopeStack scope_stack_;

  private:
    Boolean trace_namespace_regex;
    NamespaceMapping ns_mapping_;
    NamespaceMapping const& ns_mapping;
  };


  // Usual namespace mapping.
  //
  struct Namespace : Traversal::Namespace
  {
    Namespace (Context& c, Boolean set_scope = false)
        : ctx_ (c), set_scope_ (set_scope)
    {
    }

    virtual Void
    pre (Type&);

    virtual Void
    post (Type&);

  private:
    Context& ctx_;
    Boolean set_scope_;
  };


  //
  //
  template <typename X>
  struct Has : X
  {
    Has (Boolean& result)
        : result_ (result)
    {
    }

    virtual Void
    traverse (typename X::Type&)
    {
      result_ = true;
    }

  private:
    Boolean& result_;
  };

  // Checks if scope 'Y' names any of 'X'
  //
  template <typename X, typename Y>
  Boolean
  has (Y& y)
  {
    Boolean result (false);

    Has<X> t (result);

    Traversal::Scope s;
    Traversal::Names n;
    s.edge_traverser (n);
    n.node_traverser (t);

    s.traverse (y);

    return result;
  }


  // Fundamental type mapping helper.
  //
  struct Fundamental : Traversal::Fundamental::Type,
                       Traversal::Fundamental::String,
                       Traversal::Fundamental::NormalizedString,
                       Traversal::Fundamental::Token,
                       Traversal::Fundamental::Name,
                       Traversal::Fundamental::NameToken,
                       Traversal::Fundamental::NCName,
                       Traversal::Fundamental::Id,
                       Traversal::Fundamental::IdRef
  {
    virtual Void
    fundamental_type (SemanticGraph::Fundamental::Type& t) = 0;

    virtual Void
    fundamental_template (SemanticGraph::Fundamental::Type& t) = 0;

    virtual Void
    traverse (SemanticGraph::Fundamental::Type& t)
    {
      fundamental_type (t);
    }

    virtual Void
    traverse (SemanticGraph::Fundamental::String& t)
    {
      fundamental_template (t);
    }

    virtual Void
    traverse (SemanticGraph::Fundamental::NormalizedString& t)
    {
      fundamental_template (t);
    }

    virtual Void
    traverse (SemanticGraph::Fundamental::Token& t)
    {
      fundamental_template (t);
    }

    virtual Void
    traverse (SemanticGraph::Fundamental::Name& t)
    {
      fundamental_template (t);
    }

    virtual Void
    traverse (SemanticGraph::Fundamental::NameToken& t)
    {
      fundamental_template (t);
    }

    virtual Void
    traverse (SemanticGraph::Fundamental::NCName& t)
    {
      fundamental_template (t);
    }

    virtual Void
    traverse (SemanticGraph::Fundamental::Id& t)
    {
      fundamental_template (t);
    }

    virtual Void
    traverse (SemanticGraph::Fundamental::IdRef& t)
    {
      fundamental_template (t);
    }
  };
}

#endif  // CXX_TREE_ELEMENTS_HXX
