// file      : xsd/cxx/parser/elements.hxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2006 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#ifndef CXX_PARSER_ELEMENTS_HXX
#define CXX_PARSER_ELEMENTS_HXX

#include <sstream>

#include <cxx/elements.hxx>

#include <cxx/parser/cli.hxx>

namespace CXX
{
  namespace Parser
  {
    //
    //
    class Context: public CXX::Context
    {
    public:
      Context (std::wostream& o, CLI::Options const& ops)
          : CXX::Context (o,
                          ops.value<CLI::char_type> (),
                          ops.value<CLI::include_with_brackets> (),
                          ops.value<CLI::include_prefix> (),
                          ops.value<CLI::export_symbol> (),
                          ops.value<CLI::namespace_regex> (),
                          ops.value<CLI::namespace_regex_trace> (),
                          ops.value<CLI::generate_inline> ()),
            xml_parser (xml_parser_),
            parser_template (parser_template_),
            xml_parser_ (ops.value<CLI::xml_parser> ()),
            parser_template_ (L"::xsd::cxx::parser::"
                              + xml_parser
                              + L"::traverser")
      {
      }

    protected:
      Context (Context& c)
          : CXX::Context (c),
            xml_parser (c.xml_parser),
            parser_template (c.parser_template)
      {
      }

      Context (Context& c, std::wostream& o)
          : CXX::Context (c, o),
            xml_parser (c.xml_parser),
            parser_template (c.parser_template)
      {
      }

    public:
      String& xml_parser;
      String& parser_template;

    private:
      String xml_parser_;
      String parser_template_;
    };


    //
    //
    struct TemplateParamMemberDecl : Traversal::Member,
                                     protected virtual Context
    {
      TemplateParamMemberDecl (Context& c)
          : Context (c)
      {
      }

      virtual Void
      traverse (Type& m)
      {
        if (!skip (m))
          os << ", typename _xsd_" << ename (m);
      }
    };


    //
    //
    struct TemplateParamBaseDecl : Traversal::Complex,
                                   protected virtual Context
    {
      TemplateParamBaseDecl (Context& c)
          : Context (c), member_ (c)
      {
        *this >> inherits_ >> *this;
        *this >> names_ >> member_;
      }

      virtual Void
      traverse (Type& c)
      {
        String const& name (ename (c));

        os << ", typename _xsd_" << name;

        enter_scope (name);
        Traversal::Complex::traverse (c);
        leave_scope ();
      }

    private:
      Traversal::Inherits inherits_;

      TemplateParamMemberDecl member_;
      Traversal::Names names_;
    };


    //
    //
    struct TemplateParamMemberUse : Traversal::Member,
                                    protected virtual Context
    {
      TemplateParamMemberUse (Context& c)
          : Context (c)
      {
      }

      virtual Void
      traverse (Type& m)
      {
        if (!skip (m))
          os << ", _xsd_" << ename (m);
      }
    };


    //
    //
    struct TemplateParamBaseUse : Traversal::Complex,
                                  protected virtual Context
    {
      TemplateParamBaseUse (Context& c)
          : Context (c), member_ (c)
      {
        *this >> inherits_ >> *this;
        *this >> names_ >> member_;
      }

      virtual Void
      traverse (Type& c)
      {
        String const& name (ename (c));

        os << ", _xsd_" << name;

        enter_scope (name);
        Traversal::Complex::traverse (c);
        leave_scope ();
      }

    private:
      Traversal::Inherits inherits_;

      TemplateParamMemberUse member_;
      Traversal::Names names_;
    };


    //
    //
    struct TemplateParamDeclList: Traversal::Complex,
                                  protected virtual Context
    {
      TemplateParamDeclList (Context& c)
          : Context (c),
            template_param_member_ (c),
            template_param_base_ (c)
      {
        template_param_names_ >> template_param_member_;
        template_param_inherits_ >> template_param_base_;
      }

      virtual Void
      traverse (Type& c)
      {
        String const& name (ename (c));

        os << "typename _xsd_" << name;

        enter_scope (name);

        inherits (c, template_param_inherits_);
        names (c, template_param_names_);

        leave_scope ();
      }

    private:
      TemplateParamMemberDecl template_param_member_;
      Traversal::Names template_param_names_;

      TemplateParamBaseDecl template_param_base_;
      Traversal::Inherits template_param_inherits_;
    };


    //
    //
    struct TemplateParamUseList: Traversal::Complex,
                                 protected virtual Context
    {
      TemplateParamUseList (Context& c)
          : Context (c),
            template_param_member_ (c),
            template_param_base_ (c)
      {
        template_param_names_ >> template_param_member_;
        template_param_inherits_ >> template_param_base_;
      }

      virtual Void
      traverse (Type& c)
      {
        String const& name (ename (c));

        os << "_xsd_" << name;

        enter_scope (name);

        inherits (c, template_param_inherits_);
        names (c, template_param_names_);

        leave_scope ();
      }

    private:
      TemplateParamMemberUse template_param_member_;
      Traversal::Names template_param_names_;

      TemplateParamBaseUse template_param_base_;
      Traversal::Inherits template_param_inherits_;
    };


    //
    //
    struct ParserParamDecl : Traversal::Complex,
                             Traversal::Member,
                             protected virtual Context
    {
      ParserParamDecl (Context& c, Boolean name_arg)
          : Context (c), first_ (true), name_arg_ (name_arg)
      {
        *this >> inherits_ >> *this;
        *this >> names_ >> *this;
      }

      virtual Void
      traverse (SemanticGraph::Complex& c)
      {
        enter_scope (ename (c));
        Traversal::Complex::traverse (c);
        leave_scope ();
      }

      virtual Void
      traverse (SemanticGraph::Member& m)
      {
        if (skip (m)) return;

        String const& name (ename (m));

        if (!first_)
          os << ", ";
        else
          first_ = false;

        os << parser_template << "< _xsd_" << name << ", " <<
          char_type << " >&";

        if (name_arg_)
          os << " " << name;
      }

    private:
      Traversal::Inherits inherits_;
      Traversal::Names names_;

      Boolean first_;
      Boolean name_arg_;
    };


    //
    //
    struct Base : Traversal::Complex,
                  protected virtual Context
    {
      Base (Context& c)
          : Context (c), base_ (c), member_ (c)
      {
        *this >> inherits_ >> base_;
        *this >> names_ >> member_;
      }

      virtual Void
      traverse (Type& c)
      {
        String const& name (ename (c));
        String type (fq_name (c));

        os << type << "< _xsd_" << name;

        enter_scope (name);
        Traversal::Complex::traverse (c);
        leave_scope ();

        os << " >";
      }

    private:
      TemplateParamBaseUse base_;
      Traversal::Inherits inherits_;


      TemplateParamMemberUse member_;
      Traversal::Names names_;
    };
  }
}

#endif  // CXX_PARSER_ELEMENTS_HXX
