// file      : xsd/cxx/parser/generator.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2006 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <cxx/parser/generator.hxx>


#include <cxx/parser/elements.hxx>
#include <cxx/parser/parser-header.hxx>
#include <cxx/parser/parser-inline.hxx>
#include <cxx/parser/parser-template.hxx>

#include <xsd-frontend/semantic-graph.hxx>

#include <backend-elements/regex.hxx>
#include <backend-elements/indentation/cxx.hxx>
#include <backend-elements/indentation/sloc.hxx>
#include <backend-elements/indentation/clip.hxx>

#include <cult/containers/set.hxx>

#include <boost/filesystem/fstream.hpp>

#include <cctype>   // std::toupper
#include <iostream>
#include <cstdio>   // std::remove

#include "../../usage.hxx"

using std::endl;
using std::wcerr;

typedef
boost::filesystem::wifstream
WideInputFileStream;

typedef
boost::filesystem::wofstream
WideOutputFileStream;

using namespace XSDFrontend::SemanticGraph;

struct AutoUnlink
{
  AutoUnlink (Path const& file, WideOutputFileStream& os)
      : file_ (file), os_ (os), canceled_ (false)
  {
  }

  ~AutoUnlink ()
  {
    if (!canceled_)
    {
      os_.close ();
      std::remove (file_.native_file_string ().c_str ());
    }
  }

  void
  cancel ()
  {
    canceled_ = true;
  }

private:
  Path file_;
  WideOutputFileStream& os_;
  bool canceled_;
};

namespace CXX
{
  namespace
  {
    Char const copyright[] =
    "// Copyright (C) 2005-2006 Code Synthesis Tools CC\n"
    "//\n"
    "// This program was generated by XML Schema Definition Compiler (XSD)\n"
    "//\n"
    "// This program is free software; you can redistribute it and/or modify\n"
    "// it under the terms of the GNU General Public License version 2 as\n"
    "// published by the Free Software Foundation.\n"
    "//\n"
    "// This program is distributed in the hope that it will be useful,\n"
    "// but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
    "// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
    "// GNU General Public License for more details.\n"
    "//\n"
    "// You should have received a copy of the GNU General Public License\n"
    "// along with this program; if not, write to the Free Software\n"
    "// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA\n"
    "//\n"
    "// In addition, as a special exception, Code Synthesis Tools CC gives\n"
    "// permission to link this program with the Xerces-C++ library (or with\n"
    "// modified versions of Xerces-C++ that use the same license as Xerces-C++),\n"
    "// and distribute linked combinations including the two. You must obey\n"
    "// the GNU General Public License version 2 in all respects for all of\n"
    "// the code used other than Xerces-C++. If you modify this copy of the\n"
    "// program, you may extend this exception to your version of the program,\n"
    "// but you are not obligated to do so. If you do not wish to do so, delete\n"
    "// this exception statement from your version.\n"
    "//\n"
    "// In addition, Code Synthesis Tools CC makes a special exception for\n"
    "// the Free/Libre and Open Source Software (FLOSS) which is described\n"
    "// in the accompanying FLOSSE file.\n"
    "//\n\n";
  }

  namespace Parser
  {
    namespace CLI
    {
      extern Key char_type                = "char-type";
      extern Key output_dir               = "output-dir";
      extern Key xml_parser               = "xml-parser";
      extern Key generate_inline          = "generate-inline";
      extern Key namespace_regex          = "namespace-regex";
      extern Key namespace_regex_trace    = "namespace-regex-trace";
      extern Key include_with_brackets    = "include-with-brackets";
      extern Key include_prefix           = "include-prefix";
      extern Key hxx_suffix               = "hxx-suffix";
      extern Key ixx_suffix               = "ixx-suffix";
      extern Key txx_suffix               = "txx-suffix";
      extern Key hxx_regex                = "hxx-regex";
      extern Key ixx_regex                = "ixx-regex";
      extern Key txx_regex                = "txx-regex";
      extern Key hxx_prologue             = "hxx-prologue";
      extern Key ixx_prologue             = "ixx-prologue";
      extern Key txx_prologue             = "txx-prologue";
      extern Key prologue                 = "prologue";
      extern Key hxx_epilogue             = "hxx-epilogue";
      extern Key ixx_epilogue             = "ixx-epilogue";
      extern Key txx_epilogue             = "txx-epilogue";
      extern Key epilogue                 = "epilogue";
      extern Key hxx_prologue_file        = "hxx-prologue-file";
      extern Key ixx_prologue_file        = "ixx-prologue-file";
      extern Key txx_prologue_file        = "txx-prologue-file";
      extern Key prologue_file            = "prologue-file";
      extern Key hxx_epilogue_file        = "hxx-epilogue-file";
      extern Key ixx_epilogue_file        = "ixx-epilogue-file";
      extern Key txx_epilogue_file        = "txx-epilogue-file";
      extern Key epilogue_file            = "epilogue-file";
      extern Key export_symbol            = "export-symbol";
      extern Key show_sloc                = "show-sloc";
    }
  }

  Void Parser::Generator::
  usage (NarrowString const& cmd)
  {
    std::wostream& e (wcerr);
    ::CLI::Indent::Clip< ::CLI::OptionsUsage, WideChar> clip (e);

    e << "--char-type <type>" << endl
      << " Use <type> as the base character type. Valid\n"
      << " values are 'char' (default) and 'wchar_t'."
      << endl;

    e << "--output-dir <dir>" << endl
      << " Write generated files to <dir> instead of current\n"
      << " directory."
      << endl;

    e << "--xml-parser <parser>" << endl
      << " Use <parser> as the underlying XML parser. Valid\n"
      << " values are 'xerces'" <<
      (cmd == "cxx-parser" ? " (default)" : "") << " and 'expat'" <<
      (cmd == "cxx-parser-e" ? " (default)" : "") << "."
      << endl;


    e << "--generate-inline" << endl
      << " Generate certain functions inline."
      << endl;

    e << "--namespace-regex <regex>" << endl
      << " Add the provided regular expression to the list of\n"
      << " regular expressions used to translate XML Schema\n"
      << " namespace names to C++ namespace names. The first\n"
      << " successful substitution is used. The last provided\n"
      << " expression is considered first. Expression is a\n"
      << " perl-like regex in the form /pattern/replacement/."
      << endl;

    e << "--namespace-regex-trace" << endl
      << " Trace the process of applying regular expressions\n"
      << " specified with the --namespace-regex option."
      << endl;

    e << "--include-with-brackets" << endl
      << " Use angle brackets (<>) instead of quotes (\"\") in\n"
      << " generated #include directives for included and\n"
      << " imported schema definitions."
      << endl;

    e << "--include-prefix <prefix>" << endl
      << " Add <prefix> to generated #include directive paths\n"
      << " for included and imported schema definitions."
      << endl;

    e << "--hxx-suffix <suffix>" << endl
      << " Use <suffix> instead of the default '.hxx' to\n"
      << " construct the name of the header file."
      << endl;

    e << "--ixx-suffix <suffix>" << endl
      << " Use <suffix> instead of the default '.ixx' to\n"
      << " construct the name of the inline file."
      << endl;

    e << "--txx-suffix <suffix>" << endl
      << " Use <suffix> instead of the default '.txx' to\n"
      << " construct the name of the template file."
      << endl;

    e << "--hxx-regex <regex>" << endl
      << " Use <regex> to construct the name of the header\n"
      << " file. Expression is a perl regex in the form\n"
      << " /pattern/replacement/."
      << endl;

    e << "--ixx-regex <regex>" << endl
      << " Use <regex> to construct the name of the inline\n"
      << " file. Expression is a perl regex in the form\n"
      << " /pattern/replacement/."
      << endl;

    e << "--txx-regex <regex>" << endl
      << " Use <regex> to construct the name of the template\n"
      << " file. Expression is a perl regex in the form\n"
      << " /pattern/replacement/."
      << endl;


    // Prologues.
    //
    e << "--hxx-prologue <text>" << endl
      << " Insert <text> at the beginning of the header file."
      << endl;

    e << "--ixx-prologue <text>" << endl
      << " Insert <text> at the beginning of the inline file."
      << endl;

    e << "--txx-prologue <text>" << endl
      << " Insert <text> at the beginning of the template\n"
      << " file."
      << endl;

    e << "--prologue <text>" << endl
      << " Insert <text> at the beginning of each generated\n"
      << " file for which there is no file-specific prologue."
      << endl;


    // Epilogues.
    //
    e << "--hxx-epilogue <text>" << endl
      << " Insert <text> at the end of the header file."
      << endl;

    e << "--ixx-epilogue <text>" << endl
      << " Insert <text> at the end of the inline file."
      << endl;

    e << "--txx-epilogue <text>" << endl
      << " Insert <text> at the end of the template file."
      << endl;

    e << "--epilogue <text>" << endl
      << " Insert <text> at the end of each generated file\n"
      << " for which there is no file-specific epilogue."
      << endl;


    // Prologue files.
    //
    e << "--hxx-prologue-file <file>" << endl
      << " Insert the content of the <file> at the beginning\n"
      << " of the header file."
      << endl;

    e << "--ixx-prologue-file <file>" << endl
      << " Insert the content of the <file> at the beginning\n"
      << " of the inline file."
      << endl;

    e << "--txx-prologue-file <file>" << endl
      << " Insert the content of the <file> at the beginning\n"
      << " of the template file."
      << endl;

    e << "--prologue-file <file>" << endl
      << " Insert the content of the <file> at the beginning\n"
      << " of each generated file for which there is no file-\n"
      << " specific prologue file."
      << endl;


    // Epilogue files.
    //
    e << "--hxx-epilogue-file <file>" << endl
      << " Insert the content of the <file> at the end of\n"
      << " the header file."
      << endl;

    e << "--ixx-epilogue-file <file>" << endl
      << " Insert the content of the <file> at the end of\n"
      << " the inline file."
      << endl;

    e << "--txx-epilogue-file <file>" << endl
      << " Insert the content of the <file> at the end of\n"
      << " the template file."
      << endl;

    e << "--epilogue-file <file>" << endl
      << " Insert the content of the <file> at the end of\n"
      << " each generated file for which there is no file-\n"
      << " specific epilogue file."
      << endl;

    e << "--export-symbol <symbol>" << endl
      << " Export symbol for Win32 DLL export/import control."
      << endl;

    e << "--show-sloc" << endl
      << " Show the number of generated physical source lines\n"
      << " of code (SLOC)."
      << endl;
  }

  Parser::CLI::OptionsSpec Parser::Generator::
  options_spec (NarrowString const& cmd)
  {
    CLI::OptionsSpec spec;

    spec.option<CLI::char_type> ().default_value ("char");
    spec.option<CLI::xml_parser> ().default_value (
      cmd == "cxx-parser-e" ? "expat" : "xerces");

    spec.option<CLI::hxx_suffix> ().default_value (".hxx");
    spec.option<CLI::ixx_suffix> ().default_value (".ixx");
    spec.option<CLI::txx_suffix> ().default_value (".txx");

    return spec;
  }


  namespace
  {
    Void
    append (WideOutputFileStream& os,
            NarrowString const& path,
            WideInputFileStream& default_is)
    {
      using std::ios_base;

      if (path)
      {
        WideInputFileStream is (path, ios_base::in | ios_base::binary);

        if (!is.is_open ())
        {
          wcerr << path.c_str () << ": error: unable to open in read mode"
                << endl;

          throw Parser::Generator::Failed ();
        }

        os << is.rdbuf ();

      }
      else if (default_is.is_open ())
      {
        os << default_is.rdbuf ();
        default_is.seekg (0, ios_base::beg);
      }
    }

    Void
    append (WideOutputFileStream& os,
            Cult::Containers::Vector<NarrowString> const& primary,
            Cult::Containers::Vector<NarrowString> const& def)
    {
      Cult::Containers::Vector<NarrowString> const& v (
        primary.empty () ? def : primary);

      for (Containers::Vector<NarrowString>::ConstIterator
             i (v.begin ()), e (v.end ()); i != e; ++i)
      {
        os << i->c_str () << endl;
      }
    }
  }


  Void Parser::Generator::
  generate (Parser::CLI::Options const& ops,
            Schema& schema,
            Path const& file_path)
  {
    using std::ios_base;
    using BackendElements::Regex::perl_s;
    namespace Indentation = BackendElements::Indentation;

    NarrowString name (file_path.leaf ());

    NarrowString hxx_suffix (ops.value <CLI::hxx_suffix> ());
    NarrowString ixx_suffix (ops.value <CLI::ixx_suffix> ());
    NarrowString txx_suffix (ops.value <CLI::txx_suffix> ());

    NarrowString hxx_expr (ops.value <CLI::hxx_regex> ().empty ()
                           ? "#^(.+?)(\\.[^./\\\\]+)?$#$1" + hxx_suffix + "#"
                           : ops.value <CLI::hxx_regex> ());

    NarrowString ixx_expr (ops.value <CLI::ixx_regex> ().empty ()
                           ? "#^(.+?)(\\.[^./\\\\]+)?$#$1" + ixx_suffix + "#"
                           : ops.value <CLI::ixx_regex> ());

    NarrowString txx_expr (ops.value <CLI::txx_regex> ().empty ()
                           ? "#^(.+?)(\\.[^./\\\\]+)?$#$1" + txx_suffix + "#"
                           : ops.value <CLI::txx_regex> ());


    NarrowString hxx_name (perl_s (name, hxx_expr));
    NarrowString ixx_name (perl_s (name, ixx_expr));
    NarrowString txx_name (perl_s (name, txx_expr));

    Path hxx_path (hxx_name);
    Path ixx_path (ixx_name);
    Path txx_path (txx_name);

    if (NarrowString dir  = ops.value<CLI::output_dir> ())
    {
      try
      {
        Path path (dir, boost::filesystem::native);

        hxx_path = path / hxx_path;
        ixx_path = path / ixx_path;
        txx_path = path / txx_path;
      }
      catch (InvalidPath const&)
      {
        wcerr << dir.c_str () << ": error: invalid path" << endl;
        throw Failed ();
      }
    }

    WideOutputFileStream hxx (hxx_path, ios_base::out);
    WideOutputFileStream ixx;
    WideOutputFileStream txx (txx_path, ios_base::out);

    if (!hxx.is_open ())
    {
      wcerr << hxx_path << ": error: unable to open in write mode" << endl;

      throw Failed ();
    }

    AutoUnlink unlink_hxx (hxx_path, hxx);


    Boolean inline_ (ops.value<CLI::generate_inline> ());

    if (inline_)
    {
      ixx.open (ixx_path, ios_base::out);

      if (!ixx.is_open ())
      {
        wcerr << ixx_path << ": error: unable to open in write mode" << endl;

        throw Failed ();
      }
    }

    AutoUnlink unlink_ixx (ixx_path, ixx);

    if (!inline_)
      unlink_ixx.cancel ();


    if (!txx.is_open ())
    {
      wcerr << txx_path << ": error: unable to open in write mode" << endl;

      throw Failed ();
    }

    AutoUnlink unlink_txx (txx_path, txx);


    // Print copyright and license.
    //
    hxx << copyright;
    if (inline_)
      ixx << copyright;
    txx << copyright;


    //
    //
    Context::NamespaceMapping nsm;

    // Default mapping.
    //
    nsm.push_back (L"#^.* (.*?/)??"L"(([a-zA-Z_]\\w*)(/[a-zA-Z_]\\w*)*)$#$2#");
    nsm.push_back (L"#^XMLSchema.xsd http://www.w3.org/2001/XMLSchema$#xml_schema#");

    // Custom mappings.
    //
    for (Containers::Vector<NarrowString>::ConstIterator
           i (ops.value<CLI::namespace_regex> ().begin ()),
           e (ops.value<CLI::namespace_regex> ().end ()); i != e; ++i)
    {
      nsm.push_back (*i);
    }


    // Prologue.
    //
    WideInputFileStream prologue;
    {
      NarrowString name (ops.value<CLI::prologue_file> ());

      if (!name.empty ())
      {
        prologue.open (name, ios_base::in | ios_base::binary);

        if (!prologue.is_open ())
        {
          wcerr << name.c_str () << ": error: unable to open in read mode"
                << endl;

          throw Failed ();
        }
      }
    }


    // Epilogue.
    //
    WideInputFileStream epilogue;
    {
      NarrowString name (ops.value<CLI::epilogue_file> ());

      if (!name.empty ())
      {
        epilogue.open (name, ios_base::in | ios_base::binary);

        if (!epilogue.is_open ())
        {
          wcerr << name.c_str () << ": error: unable to open in read mode"
                << endl;

          throw Failed ();
        }
      }
    }

    // SLOC counter.
    //
    UnsignedLong sloc (0);
    Boolean show_sloc (ops.value<CLI::show_sloc> ());

    try
    {

      // HXX
      //
      {
        Context ctx (hxx, ops);

        Indentation::Clip<Indentation::SLOC, WideChar> hxx_sloc (hxx);

        String guard (hxx_name);
        guard = perl_s (guard, L"/([a-z])([A-Z])/$1_$2/"); // Split words.
        std::transform (guard.begin (), guard.end(), guard.begin (), upcase);
        guard = ctx.escape (guard); // Make it a C++ id.

        hxx << "#ifndef " << guard << endl
            << "#define " << guard << endl
            << endl;


        // Copy prologue.
        //
        hxx << "// Begin prologue." << endl
            << "//" << endl;

        append (
          hxx, ops.value<CLI::hxx_prologue> (), ops.value<CLI::prologue> ());
        append (hxx, ops.value<CLI::hxx_prologue_file> (), prologue);

        hxx << "//" << endl
            << "// End prologue." << endl
            << endl;

        {
          hxx << "#include <xsd/cxx/pre.hxx>" << endl
              << endl;

          // Set auto-indentation.
          //
          Indentation::Clip<Indentation::CXX, WideChar> hxx_clip (hxx);


          // Generate.
          //
          generate_parser_header (ctx, schema, hxx_expr); //@@ move expr to ctx


          if (inline_)
            hxx << "#include \"" << ixx_name.c_str () << "\"" << endl;

          hxx << "#include \"" << txx_name.c_str () << "\"" << endl
              << endl;

          hxx << "#include <xsd/cxx/post.hxx>" << endl
              << endl;
        }

        // Copy epilogue.
        //
        hxx << "// Begin epilogue." << endl
            << "//" << endl;

        append (hxx, ops.value<CLI::hxx_epilogue_file> (), epilogue);
        append (
          hxx, ops.value<CLI::hxx_epilogue> (), ops.value<CLI::epilogue> ());

        hxx << "//" << endl
            << "// End epilogue." << endl
            << endl;

        hxx << "#endif // " << guard << endl;

        if (show_sloc)
        {
          wcerr << hxx_path << ": "
                << hxx_sloc.buffer ().count () << std::endl;

          sloc += hxx_sloc.buffer ().count ();
        }
      }


      // IXX
      //
      if (inline_)
      {
        Context ctx (ixx, ops);

        Indentation::Clip<Indentation::SLOC, WideChar> ixx_sloc (ixx);


        // Copy prologue.
        //
        ixx << "// Begin prologue." << endl
            << "//" << endl;

        append (
          ixx, ops.value<CLI::ixx_prologue> (), ops.value<CLI::prologue> ());
        append (ixx, ops.value<CLI::ixx_prologue_file> (), prologue);

        ixx << "//" << endl
            << "// End prologue." << endl
            << endl;

        {
          // Set auto-indentation.
          //
          Indentation::Clip<Indentation::CXX, WideChar> ixx_clip (ixx);


          // Generate.
          //
          generate_parser_inline (ctx, schema);
        }

        // Copy epilogue.
        //
        ixx << "// Begin epilogue." << endl
            << "//" << endl;

        append (ixx, ops.value<CLI::ixx_epilogue_file> (), epilogue);
        append (
          ixx, ops.value<CLI::ixx_epilogue> (), ops.value<CLI::epilogue> ());

        ixx << "//" << endl
            << "// End epilogue." << endl
            << endl;

        if (show_sloc)
        {
          wcerr << ixx_path << ": "
                << ixx_sloc.buffer ().count () << std::endl;

          sloc += ixx_sloc.buffer ().count ();
        }
      }


      // TXX
      //
      {
        Context ctx (txx, ops);

        Indentation::Clip<Indentation::SLOC, WideChar> txx_sloc (txx);

        // Copy prologue.
        //
        txx << "// Begin prologue." << endl
            << "//" << endl;

        append (
          txx, ops.value<CLI::txx_prologue> (), ops.value<CLI::prologue> ());
        append (txx, ops.value<CLI::txx_prologue_file> (), prologue);

        txx << "//" << endl
            << "// End prologue." << endl
            << endl;

        {
          // Set auto-indentation.
          //
          Indentation::Clip<Indentation::CXX, WideChar> txx_clip (txx);

          if (!inline_)
            generate_parser_inline (ctx, schema);

          generate_parser_template (ctx, schema);
        }

        // Copy epilogue.
        //
        txx << "// Begin epilogue." << endl
            << "//" << endl;

        append (txx, ops.value<CLI::txx_epilogue_file> (), epilogue);
        append (
          txx, ops.value<CLI::txx_epilogue> (), ops.value<CLI::epilogue> ());

        txx << "//" << endl
            << "// End epilogue." << endl
            << endl;

        if (show_sloc)
        {
          wcerr << txx_path << ": "
                << txx_sloc.buffer ().count () << std::endl;

          sloc += txx_sloc.buffer ().count ();
        }

      }

      unlink_hxx.cancel ();
      unlink_ixx.cancel ();
      unlink_txx.cancel ();

      if (show_sloc)
        wcerr << "total: " << sloc << std::endl;

    }
    catch (NoNamespaceMapping const& e)
    {
      wcerr << e.file () << ":" << e.line () << ":" << e.column ()
            << ": error: unable to map XML Schema namespace '" << e.ns ()
            << "' to C++ namespace" << endl;

      wcerr << e.file () << ":" << e.line () << ":" << e.column ()
            << ": info: use --namespace-regex to provide custom mapping"
            << endl;

      throw Failed ();
    }
  }
}



